@@ -3,7 +3,7 @@ from __future__ import annotations
 from collections import deque
 from contextlib import contextmanager
 from functools import cached_property
-from typing import Any, ContextManager, Dict, Hashable, Iterator, Mapping, Optional, TYPE_CHECKING, Type, Union
+from typing import Any, ContextManager, Dict, Hashable, Iterator, Mapping, Optional, TYPE_CHECKING, Tuple, Type, Union
 from uuid import uuid4
 
 from jschon.exceptions import JSONSchemaError
@@ -17,7 +17,7 @@ if TYPE_CHECKING:
 
 __all__ = [
     'JSONSchema',
-    'Scope',
+    'Result',
 ]
 
 
@@ -29,7 +29,7 @@ class JSONSchema(JSON):
             value: Union[bool, Mapping[str, JSONCompatible]],
             *,
             catalog: Union[str, Catalog] = 'catalog',
-            session: Hashable = 'default',
+            cacheid: Hashable = 'default',
             uri: URI = None,
             metaschema_uri: URI = None,
             parent: JSON = None,
@@ -39,9 +39,8 @@ class JSONSchema(JSON):
         schema-compatible `value`.
 
         :param value: a schema-compatible Python object
-        :param catalog: the catalog in which the schema is cached
-        :param session: a session identifier, identifying which session
-            cache to put the schema in
+        :param catalog: catalog instance or catalog name
+        :param cacheid: schema cache identifier
         :param uri: the URI identifying the schema; an ``"$id"`` keyword
             appearing in `value` will override this
         :param metaschema_uri: the URI identifying the schema's metaschema;
@@ -59,12 +58,11 @@ class JSONSchema(JSON):
         self.catalog: Catalog = catalog
         """The catalog in which the schema is cached."""
 
-        self.session: Hashable = session
-        """A session identifier, identifying which session
-        cache to use for the schema."""
+        self.cacheid: Hashable = cacheid
+        """Schema cache identifier."""
 
         if uri is not None:
-            catalog.add_schema(uri, self, session=session)
+            catalog.add_schema(uri, self, cacheid=cacheid)
 
         self._uri: Optional[URI] = uri
         self._metaschema_uri: Optional[URI] = metaschema_uri
@@ -112,7 +110,7 @@ class JSONSchema(JSON):
             kwclasses = {
                 key: kwclass for key in value
                 if (key not in self.keywords and  # skip bootstrapped keywords
-                    (kwclass := self.metaschema.kwclasses.get(key)))
+                    (kwclass := self.metaschema.get_kwclass(key)))
             }
 
             for kwclass in self._resolve_dependencies(kwclasses):
@@ -157,7 +155,7 @@ class JSONSchema(JSON):
     @staticmethod
     def _resolve_dependencies(kwclasses: Dict[str, KeywordClass]) -> Iterator[KeywordClass]:
         dependencies = {
-            kwclass: [depclass for dep in kwclass.depends
+            kwclass: [depclass for dep in kwclass.depends_on
                       if (depclass := kwclasses.get(dep))]
             for kwclass in kwclasses.values()
         }
@@ -173,42 +171,40 @@ class JSONSchema(JSON):
                     yield kwclass
                     break
 
-    def validate(self) -> Scope:
+    def validate(self) -> Result:
         """Validate the schema against its metaschema."""
         return self.metaschema.evaluate(self)
 
-    def evaluate(self, instance: JSON, scope: Scope = None) -> Scope:
-        """Evaluate a JSON document.
-        
-        The returned :class:`Scope` represents the complete evaluation
-        result tree for this (sub)schema node.
+    def evaluate(self, instance: JSON, result: Result = None) -> Result:
+        """Evaluate a JSON document and return the evaluation result.
 
         :param instance: the JSON document to evaluate
-        :param scope: the dynamic evaluation scope; used by keywords
+        :param result: the current result node; given by keywords
             when invoking this method recursively
         """
-        if scope is None:
-            scope = Scope(self, instance)
+        if result is None:
+            result = Result(self, instance)
 
         if self.data is True:
             pass
 
         elif self.data is False:
-            scope.fail("The instance is disallowed by a boolean false schema")
+            result.fail("The instance is disallowed by a boolean false schema")
 
         else:
             for key, keyword in self.keywords.items():
                 if not keyword.static and keyword.can_evaluate(instance):
-                    with scope(instance, key, self) as subscope:
-                        keyword.evaluate(instance, subscope)
+                    with result(instance, key, self) as subresult:
+                        keyword.evaluate(instance, subresult)
 
             if any(
                     not child.passed
-                    for child in scope.iter_children(instance)
+                    for child in result.children.values()
+                    if child.instance.path == instance.path
             ):
-                scope.fail()
+                result.fail()
 
-        return scope
+        return result
 
     @cached_property
     def parentschema(self) -> Optional[JSONSchema]:
@@ -222,21 +218,13 @@ class JSONSchema(JSON):
                 return parent
             parent = parent.parent
 
-    @property
+    @cached_property
     def metaschema(self) -> Metaschema:
         """The schema's :class:`~jschon.vocabulary.Metaschema`."""
-        from jschon.vocabulary import Metaschema
-
         if (uri := self.metaschema_uri) is None:
             raise JSONSchemaError("The schema's metaschema URI has not been set")
 
-        if not isinstance(
-                metaschema := self.catalog.get_schema(uri, session='__meta__'),
-                Metaschema,
-        ):
-            raise JSONSchemaError(f"The schema referenced by {uri} is not a metachema")
-
-        return metaschema
+        return self.catalog.get_metaschema(uri)
 
     @property
     def metaschema_uri(self) -> Optional[URI]:
@@ -278,12 +266,12 @@ class JSONSchema(JSON):
     def uri(self, value: Optional[URI]) -> None:
         if self._uri != value:
             if self._uri is not None:
-                self.catalog.del_schema(self._uri, session=self.session)
+                self.catalog.del_schema(self._uri, cacheid=self.cacheid)
 
             self._uri = value
 
             if self._uri is not None:
-                self.catalog.add_schema(self._uri, self, session=self.session)
+                self.catalog.add_schema(self._uri, self, cacheid=self.cacheid)
 
     @property
     def canonical_uri(self) -> Optional[URI]:
@@ -310,22 +298,49 @@ class JSONSchema(JSON):
                 return node._uri.copy(fragment=relpath.uri_fragment())
 
 
-class Scope:
+class Result:
+    """The result of evaluating a JSON document
+    node against a JSON schema node.
+
+    The root of a :class:`Result` tree represents
+    a complete document evaluation result.
+    """
+
     def __init__(
             self,
             schema: JSONSchema,
             instance: JSON,
             *,
-            parent: Scope = None,
+            parent: Result = None,
             key: str = None,
-    ):
+    ) -> None:
+        self.path: JSONPointer
+        """The dynamic evaluation path to the current schema node."""
+
+        self.relpath: JSONPointer
+        """The path to the current schema node relative to the evaluating (sub)schema."""
+
         self.schema: JSONSchema = schema
+        """The evaluating (sub)schema."""
+
         self.instance: JSON = instance
-        self.parent: Optional[Scope] = parent
+        """The instance under evaluation."""
+
+        self.parent: Optional[Result] = parent
+        """The parent result node."""
+
         self.key: Optional[str] = key
-        self.children: Dict[JSONPointer, Dict[str, Scope]] = {}
+        """The index of the current schema node within its dynamic parent."""
+
+        self.children: Dict[Tuple[str, JSONPointer], Result] = {}
+        """Subresults of the current result node, indexed by schema key and instance path."""
+
         self.annotation: JSONCompatible = None
-        self.error: Optional[str] = None
+        """The annotation value of the result."""
+
+        self.error: JSONCompatible = None
+        """The error value of the result."""
+
         self._valid = True
         self._assert = True
         self._discard = False
@@ -347,20 +362,19 @@ class Scope:
             key: str,
             schema: JSONSchema = None,
             *,
-            cls: Type[Scope] = None,
-    ) -> ContextManager[Scope]:
-        """Yield a subscope of the current scope, for evaluating `instance`.
+            cls: Type[Result] = None,
+    ) -> ContextManager[Result]:
+        """Yield a subresult for the evaluation of `instance`.
         Descend down the evaluation path by `key`, into `schema` if given, or
-        within the schema of the current scope otherwise.
+        within `self.schema` otherwise.
 
-        Extension keywords may provide a custom Scope class via `cls`, which
-        is applied to all nodes within the yielded subtree.
+        Extension keywords may provide a custom :class:`Result` class via `cls`,
+        which is applied to all nodes within the yielded subtree.
         """
         if schema is None:
             schema = self.schema
 
-        self.children.setdefault(instance_path := instance.path, {})
-        self.children[instance_path][key] = (child := (cls or self.__class__)(
+        self.children[key, instance.path] = (child := (cls or self.__class__)(
             schema,
             instance,
             parent=self,
@@ -371,7 +385,7 @@ class Scope:
             yield child
         finally:
             if child._discard:
-                del self.children[instance_path][key]
+                del self.children[key, instance.path]
 
     @cached_property
     def globals(self) -> Dict:
@@ -382,74 +396,70 @@ class Scope:
 
     @cached_property
     def schema_node(self) -> JSON:
-        """Return the schema node associated with this scope."""
+        """Return the current schema node."""
         return self.relpath.evaluate(self.schema)
 
-    def sibling(self, instance: JSON, key: str) -> Optional[Scope]:
+    def sibling(self, instance: JSON, key: str) -> Optional[Result]:
+        """Return a sibling schema node's evaluation result for `instance`."""
         try:
-            return self.parent.children[instance.path][key] if self.parent else None
+            return self.parent.children[key, instance.path] if self.parent else None
         except KeyError:
             return None
 
     def annotate(self, value: JSONCompatible) -> None:
-        """Set an annotation on the scope."""
+        """Annotate the result."""
         self.annotation = value
 
-    def fail(self, error: str = None) -> None:
-        """Flag the scope as invalid, optionally with an error message."""
+    def fail(self, error: JSONCompatible = None) -> None:
+        """Mark the result as invalid, optionally with an error."""
         self._valid = False
         self.error = error
 
     def pass_(self) -> None:
-        """Flag the scope as valid.
+        """Mark the result as valid.
         
-        A scope is initially valid, so this should ordinarily only need
-        to be called by a keyword when it must reverse a scope failure.
+        A result is initially valid, so this should only need
+        to be called by a keyword when it must reverse a failure.
         """
         self._valid = True
         self.error = None
 
     def noassert(self) -> None:
-        """Indicate that the scope's validity should not affect its
-        assertion result."""
+        """Indicate that evaluation passes regardless of validity."""
         self._assert = False
 
     def discard(self) -> None:
-        """Indicate that the scope should be ignored and discarded."""
+        """Indicate that the result should be ignored and discarded."""
         self._discard = True
 
     def refschema(self, schema: JSONSchema) -> None:
-        """Set the referenced schema for the scope of a by-reference keyword.
+        """Set the referenced schema for a by-reference keyword.
         
-        The referenced schema's URI is then returned by :attr:`absolute_uri`,
-        instead of calculating the absolute URI using the containing schema.
+        This ensures that :attr:`absolute_uri` returns the URI of the
+        referenced schema rather than the referencing keyword.
         """
         self._refschema = schema
 
     @property
     def valid(self) -> bool:
-        """Return the validation result of the scope.
-
-        :rtype: bool
-        """
+        """Return the validity of the instance against the schema."""
         return self._valid
 
     @property
     def passed(self) -> bool:
-        """Return the assertion result of the scope.
-        
-        In the current implementation, this can only ever differ from
-        :attr:`valid` for an "if" keyword subscope: its validation result
-        may be false (triggering "else") while its assertion result is always
-        true. For the root scope (representing the overall document evaluation
-        result), :attr:`valid` will always equal :attr:`passed`.
+        """Return the assertion result for the schema node.
 
-        :rtype: bool
+        In the standard JSON Schema vocabulary, this can only differ
+        from :attr:`valid` for the ``if`` keyword: validity may be false
+        (triggering ``else``) while its assertion result is always true.
+
+        For the root result node, :attr:`passed` will always equal :attr:`valid`.
         """
         return self._valid or not self._assert
 
     @property
     def absolute_uri(self) -> Optional[URI]:
+        """Return the absolute URI of the current schema node."""
         if self._refschema is not None:
             return self._refschema.canonical_uri
 
@@ -460,13 +470,6 @@ class Scope:
                 relpath = self.relpath
             return schema_uri.copy(fragment=relpath.uri_fragment())
 
-    def iter_children(self, instance: JSON = None) -> Iterator[Scope]:
-        """Return an iterator over child scopes of this scope, optionally
-        filtered by an instance to which they apply."""
-        for instance_path, keyword_scopes in self.children.items():
-            if instance is None or instance.path == instance_path:
-                yield from keyword_scopes.values()
-
     def collect_annotations(self, instance: JSON = None, key: str = None) -> Iterator[JSONCompatible]:
         """Return an iterator over annotations produced in this subtree,
         optionally filtered by instance and/or keyword."""
@@ -475,20 +478,34 @@ class Scope:
                     (key is None or key == self.key) and \
                     (instance is None or instance.path == self.instance.path):
                 yield self.annotation
-            for child in self.iter_children():
+            for child in self.children.values():
                 yield from child.collect_annotations(instance, key)
 
+    def collect_errors(self, instance: JSON = None, key: str = None) -> Iterator[JSONCompatible]:
+        """Return an iterator over errors produced in this subtree,
+        optionally filtered by instance and/or keyword."""
+        if not self._valid and not self._discard:
+            if self.error is not None and \
+                    (key is None or key == self.key) and \
+                    (instance is None or instance.path == self.instance.path):
+                yield self.error
+            for child in self.children.values():
+                yield from child.collect_errors(instance, key)
+
     def output(self, format: str, **kwargs: Any) -> JSONCompatible:
         """Return the evaluation result in the specified `format`.
 
-        :param format: The format of the returned document. The default
-            :class:`~jschon.output.JSONSchemaOutputFormatter` supports
-            the ``flag``, ``basic``, ``detailed`` and ``verbose`` output
-            formats, as defined by the JSON Schema specification.
-        :param kwargs: Additional keyword arguments to pass to the output
-            formatter.
+        :param format: One of the standard JSON Schema output formats --
+            ``flag``, ``basic``, ``detailed`` or ``verbose`` -- or any
+            format registered with the :func:`~jschon.output.output_formatter`
+            decorator.
+        :param kwargs: Keyword arguments to pass to the output formatter.
         """
-        return self.schema.catalog.output_formatter.create_output(self, format, **kwargs)
+        from jschon.output import create_output
+        return create_output(self, format, **kwargs)
 
     def __str__(self) -> str:
-        return str(self.path)
+        s = 'valid' if self.valid else 'invalid'
+        if self.parent:
+            s = f'{self.path}: {s}'
+        return s
