@@ -1,10 +1,10 @@
 from typing import Mapping
 
-from jschon.exceptions import JSONSchemaError, URIError, CatalogError
+from jschon.exceptions import CatalogError, JSONSchemaError, URIError
 from jschon.json import JSON
-from jschon.jsonschema import JSONSchema, Scope
+from jschon.jsonschema import JSONSchema, Result
 from jschon.uri import URI
-from jschon.vocabulary import Keyword, PropertyApplicator, Metaschema
+from jschon.vocabulary import Keyword, Metaschema, ObjectOfSubschemas
 
 __all__ = [
     'SchemaKeyword',
@@ -95,12 +95,12 @@ class RefKeyword(Keyword):
                 raise JSONSchemaError(f'No base URI against which to resolve the "$ref" value "{uri}"')
 
         self.refschema = self.parentschema.catalog.get_schema(
-            uri, metaschema_uri=self.parentschema.metaschema_uri, session=self.parentschema.session
+            uri, metaschema_uri=self.parentschema.metaschema_uri, cacheid=self.parentschema.cacheid
         )
 
-    def evaluate(self, instance: JSON, scope: Scope) -> None:
-        self.refschema.evaluate(instance, scope)
-        scope.refschema(self.refschema)
+    def evaluate(self, instance: JSON, result: Result) -> None:
+        self.refschema.evaluate(instance, result)
+        result.refschema(self.refschema)
 
 
 class AnchorKeyword(Keyword):
@@ -115,7 +115,7 @@ class AnchorKeyword(Keyword):
         else:
             raise JSONSchemaError(f'No base URI for "$anchor" value "{value}"')
 
-        parentschema.catalog.add_schema(uri, parentschema, session=parentschema.session)
+        parentschema.catalog.add_schema(uri, parentschema, cacheid=parentschema.cacheid)
 
 
 class DynamicRefKeyword(Keyword):
@@ -124,12 +124,7 @@ class DynamicRefKeyword(Keyword):
     def __init__(self, parentschema: JSONSchema, value: str):
         super().__init__(parentschema, value)
 
-        # this is not required by the spec, but it doesn't make sense
-        # for a $dynamicRef *not* to end in a plain-name fragment
-        if (fragment := URI(value).fragment) is None or '/' in fragment:
-            raise JSONSchemaError('The value for "$dynamicRef" must end in a plain-name fragment')
-
-        self.fragment = fragment
+        self.fragment = URI(value).fragment
         self.refschema = None
         self.dynamic = False
 
@@ -142,25 +137,25 @@ class DynamicRefKeyword(Keyword):
                 raise JSONSchemaError(f'No base URI against which to resolve the "$dynamicRef" value "{uri}"')
 
         self.refschema = self.parentschema.catalog.get_schema(
-            uri, metaschema_uri=self.parentschema.metaschema_uri, session=self.parentschema.session
+            uri, metaschema_uri=self.parentschema.metaschema_uri, cacheid=self.parentschema.cacheid
         )
         if (dynamic_anchor := self.refschema.get("$dynamicAnchor")) and dynamic_anchor.data == self.fragment:
             self.dynamic = True
 
-    def evaluate(self, instance: JSON, scope: Scope) -> None:
+    def evaluate(self, instance: JSON, result: Result) -> None:
         refschema = self.refschema
 
         if self.dynamic:
-            target_scope = scope
+            target = result
             checked_uris = set()
 
-            while target_scope is not None:
-                if (base_uri := target_scope.schema.base_uri) is not None and base_uri not in checked_uris:
+            while target is not None:
+                if (base_uri := target.schema.base_uri) is not None and base_uri not in checked_uris:
                     checked_uris |= {base_uri}
                     target_uri = URI(f"#{self.fragment}").resolve(base_uri)
                     try:
                         found_schema = self.parentschema.catalog.get_schema(
-                            target_uri, session=self.parentschema.session
+                            target_uri, cacheid=self.parentschema.cacheid
                         )
                         if (dynamic_anchor := found_schema.get("$dynamicAnchor")) and \
                                 dynamic_anchor.data == self.fragment:
@@ -168,10 +163,10 @@ class DynamicRefKeyword(Keyword):
                     except CatalogError:
                         pass
 
-                target_scope = target_scope.parent
+                target = target.parent
 
-        refschema.evaluate(instance, scope)
-        scope.refschema(refschema)
+        refschema.evaluate(instance, result)
+        result.refschema(refschema)
 
 
 class DynamicAnchorKeyword(Keyword):
@@ -186,10 +181,10 @@ class DynamicAnchorKeyword(Keyword):
         else:
             raise JSONSchemaError(f'No base URI for "$dynamicAnchor" value "{value}"')
 
-        parentschema.catalog.add_schema(uri, parentschema, session=parentschema.session)
+        parentschema.catalog.add_schema(uri, parentschema, cacheid=parentschema.cacheid)
 
 
-class DefsKeyword(Keyword, PropertyApplicator):
+class DefsKeyword(Keyword, ObjectOfSubschemas):
     key = "$defs"
     static = True
 
